#!python3

import os
import pytest
#from .test_imap import send_email
from ..lib.files import temp_file
from ..__main__ import parse_args, cli_execution

# Send an email with attachment to perform tests with download the attachment
#send_email()

class TestCli:

    def test_write_nofile(self):
        """
        Raise SystemExit when no -c option is given

        """
        options = parse_args(['-ui', 'dummy', '--write_config'])

        with pytest.raises(SystemExit):
            cli_execution(options)

    def test_write_config(self):
        """
        Check if config file is written correctly

        """
        options = parse_args(['-ui', 'dummy', '-c', 'test_write_config.conf', '--write_config'])

        cli_execution(options)

        assert os.path.isfile(options.reports_conf)

    def test_report_outdated_withconfig(self):
        """
        """
        options = parse_args(['-ui', 'dummy', '-c', 'test_write_config.conf',
                              '--report', 'outdated'])

        cli_execution(options)

    def test_report_outdated(self):
        """
        """
        options = parse_args(['-ui', 'dummy', '--report', 'outdated'])

        cli_execution(options)

    def test_report_outdated_detail(self):
        """
        """
        options = parse_args(['-ui', 'dummy', '--report', 'outdated', '--detail'])

        cli_execution(options)

    def test_report_outdated_detail_ping(self):
        """
        """
        options = parse_args(['-ui', 'dummy', '--report', 'outdated', '--detail', '--ping'])

        cli_execution(options)

    def test_report_email_outdated(self):
        """
        """
        options = parse_args(['-ui', 'dummy', '--report', 'email_outdated'])

        cli_execution(options)

    def test_report_inventory(self):
        """
        """
        # Read inventory file from data dir
        inventory = os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                                  '..', 'data',
                                                  'test_inventory.csv'))

        output = temp_file('inventory_test.csv')

        options = parse_args(['-ui', 'dummy', '--report', 'inventory',
                              '-i', inventory, '-o', output])

        cli_execution(options)

    def test_report_inventoryurl(self):
        """
        """
        # Read inventory file from data dir
        inventory = 'https://raw.githubusercontent.com/pablodav/burp_server_reports/master/burp_reports/data/test_inventory.csv'
        output = temp_file('inventory_test.csv')

        options = parse_args(['-ui', 'dummy', '--report', 'inventory',
                              '-i', inventory, '-o', output])

        cli_execution(options)

    def test_report_inventory_email(self):
        """
        """
        # Read inventory file from data dir
        config_file = os.path.normpath(os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '..', 'data', 'test_config.conf'))

        output = temp_file('inventory_test_email.csv')

        options = parse_args(['-ui', 'dummy', '-c', config_file, '--report',
                              'inventory', '-i', 'email_inventory', '-o', output])

        cli_execution(options)

        assert os.path.isfile(output)

    def test_demo_ui_outdated_detail(self):
        """
        """
        options = parse_args(['-ui', 'https://admin:admin@demo.burp-ui.org/api/',
                              '--report', 'outdated', '--detail'])

        cli_execution(options)

    def test_demo_ui_outdated(self):
        """
        """
        options = parse_args(['-ui', 'https://admin:admin@demo.burp-ui.org/api/',
                              '--report', 'outdated'])

        cli_execution(options)

    def test_demo_ui_outdated_debug(self):
        """
        """
        options = parse_args(['-ui', 'https://admin:admin@demo.burp-ui.org/api/',
                              '--report', 'outdated',
                              '--detail',
                              '--debug'])

        cli_execution(options)

    def test_demo_ui_debug(self):
        """
        """
        options = parse_args(['-ui', 'https://admin:admin@demo.burp-ui.org/api/',
                              '--detail',
                              '--debug'])

        cli_execution(options)

    def test_version(self):
        """
        """
        options = parse_args(['--version'])

        with pytest.raises(SystemExit):
            cli_execution(options)

    def test_noapi(self):
        """
        Raise SystemExit when no burpui_apiurl option is given
        """
        options = parse_args(['--report', 'print'])

        with pytest.raises(SystemExit):
            cli_execution(options)



#!python3

from ..dummy.burpui_api_translate_dummy import BUIClients


def test_dummy():
    clients_obj = BUIClients()
    clients_dict = clients_obj.translate_clients_stats()

    assert isinstance(clients_dict, dict)

    return clients_dict



# -*- coding: utf8 -*-
import logging
from ..lib.urlget import get_url_data
from ..defaults.default_data_structure import default_client_backup_report
from collections import defaultdict
from functools import lru_cache
import requests


class Clients:
    """
    Get clients from burpui api url.

    """

    def __init__(self, apiurl):
        """
        should be the api to connect
        like: http:/user:password@server:port/api/

        :param apiurl
        """

        self.apiurl = apiurl

        self.IsMultiAgent = self._is_multi_agent()
        self.empty_backup_report = default_client_backup_report()

    @lru_cache(maxsize=32)
    def _is_multi_agent(self):
        """

        :param self:
        :return: True/None
        """
        # Notes for future detection of multi-agent mode:
        # url /api/servers/report
        # Then we should look on the list of servers with their names.
        # https://burp-ui.readthedocs.io/en/latest/api.html#get--api-servers-report
        # And use:
        # Parameters:
        # server (str)  Which server to collect data from when in multi-agent mode

        # Comments from Ziirish:
        # You'd better test /api/servers/stats
        # It returns an empty list ([]) when there are no agents and then switch back to
        # the "standalone" mode.
        # Now if the list is not empty, you'll have something like:
        # [
        # {
        #   'alive': true,
        #   'clients': 2,
        #   'name': 'burp1',
        # },
        #  {
        #   'alive': false,
        #   'clients': 0,
        #   'name': 'burp2',
        #   },
        #  ]
        serviceurl = self.apiurl + 'servers/stats'
        burpui_servers = get_url_data(serviceurl, check_multi=True)
        return_bool = None

        if burpui_servers:
            return_bool = True

        return return_bool

    @lru_cache(maxsize=32)
    def _get_clients_stats_multi(self):
        """

        :return: return a complete list of clients for each alive server in multi-agent server.
        """

        serviceurl = self.apiurl + 'servers/stats'
        burpui_servers = get_url_data(serviceurl)

        logging.debug('burpui_servers: {}'.format(burpui_servers))

        clients_stats = []

        for i in range(len(burpui_servers)):

            # Do not get data from servers offline
            if not burpui_servers[i]['alive']:
                continue

            server = burpui_servers[i]['name']
            # Not used now: burpui_api_params = {'server': server}
            # Use http://burp-ui.readthedocs.io/en/latest/api.html#get--api-clients-(server)-stats
            # The JSON returned is:
            # {
            #   [
            #     {
            #       "last": "2015-05-17 11:40:02",
            #       "name": "client1",
            #       "state": "idle",
            #       "phase": "phase1",
            #       "percent": 12,
            #     },
            #     {
            #       "last": "never",
            #       "name": "client2",
            #       "state": "idle",
            #       "phase": "phase2",
            #       "percent": 42,
            #     }
            #   ]
            # }
            serviceurl = self.apiurl + 'clients/{}/stats'.format(server)

            logging.debug('serviceurl: {}'.format(serviceurl))

            server_clients_stats = get_url_data(serviceurl=serviceurl)
            logging.debug('server_clients_stats: {}'.format(server_clients_stats))

            # Append client to clients_stats
            for cli in range(len(server_clients_stats)):
                client_stats = server_clients_stats[cli]
                client_stats['server'] = server
                clients_stats.append(client_stats)

        return clients_stats

    @lru_cache(maxsize=32)
    def _get_clients_report_multi(self):
        """
        doc: https://burp-ui.readthedocs.io/en/latest/api.html#get--api-clients-(server)-report
        GET /api/clients/(server)/report
        Returns a global report about all the clients of a given server

        :return: return a complete list of clients for each alive server in multi-agent server.
        """
        # Get burpui servers:
        serviceurl = self.apiurl + 'servers/stats'
        burpui_servers = get_url_data(serviceurl)

        logging.debug('burpui_servers: {}'.format(burpui_servers))

        clients_reports = []

        for i in range(len(burpui_servers)):

            # Do not get data from servers offline
            if not burpui_servers[i]['alive']:
                continue

            server = burpui_servers[i]['name']
            # Not used now: burpui_api_params = {'server': server}
            # Use http://burp-ui.readthedocs.io/en/latest/api.html#get--api-clients-(server)-stats
            # GET /api/clients/(server)/report
            serviceurl = self.apiurl + 'clients/{}/report'.format(server)

            logging.debug('serviceurl: {}'.format(serviceurl))

            server_clients_reports = get_url_data(serviceurl=serviceurl)
            # From demo: {'backups': [{'number': 11, 'name': 'demo3'}, {'number': 11, 'name': 'demo4'}],
            # 'clients': [{'name': 'demo3', 'stats': {'windows': 'unknown', 'totsize': 8317913635, 'total': 540904}},
            # {'name': 'demo4', 'stats': {'windows': 'unknown', 'totsize': 8317913635, 'total': 540904}}]}

            logging.debug('server_clients_reports: {}'.format(server_clients_reports))

            # Append client to clients_reports
            for cli in range(len(server_clients_reports['clients'])):
                client_stats = server_clients_reports['clients'][cli]
                client_stats['server'] = server
                clients_reports.append(client_stats)

        return clients_reports

    @lru_cache(maxsize=32)
    def _get_backup_report_stats(self, client, number, server=None):
        """
        GET /api/client/(server)/report/(name)/(int: backup)
        GET /api/client/report/(name)/(int: backup)
        will be used: totsize, received, duration.
        """

        serviceurl = self.apiurl + 'client/report/{}/{}'.format(client, number)

        if server:
            serviceurl = self.apiurl + 'client/{}/report/{}/{}'.format(server, client, number)

        logging.debug('apiurl: {}'.format(serviceurl))

        backup_report = get_url_data(serviceurl)

        return backup_report

    @lru_cache(maxsize=32)
    def _get_client_report_stats(self, client, server=None):
        """
        https://burp-ui.readthedocs.io/en/latest/api.html#get--api-client-stats-(name)
        GET  /api/client/stats/(name)
        GET /api/client/(server)/stats/(name)

        :return list:
        [
          {
            "date": "2015-01-25 13:32:00",
            "deletable": true,
            "encrypted": true,
            "received": 123,
            "size": 1234,
            "number": 1
          },
        ]
        """

        serviceurl = self.apiurl + 'client/stats/{}'.format(client)

        if server:
            serviceurl = self.apiurl + 'client/{}/stats/{}'.format(server, client)

        logging.debug('apiurl: {}'.format(serviceurl))

        client_report = get_url_data(serviceurl, ignore_empty=True)

        return client_report

    @lru_cache(maxsize=32)
    def get_clients_stats(self):
        """
        #  server.get_all_clients()
        #
        :return list: example: [{
            "phase": 'null',
            "percent": 0,
            "state": "idle",
            "last": "2016-06-23 14:33:06-03:00",
            "name": "monitor"},
            {
                "last": "2015-05-17 11:40:02-03:00",
                "name": "client1",
                "state": "idle",
                "phase": "phase1",
                "percent": 12,
            },
            {
                "last": "never",
                "name": "client2",
                "state": "idle",
                "phase": "phase2",
                "percent": 42,
            }
        ]
        """

        logging.debug("Url received: {}".format(self.apiurl))

        if self.IsMultiAgent:
            clients_stats = self._get_clients_stats_multi()
        else:
            serviceurl = self.apiurl + 'clients/stats'
            clients_stats = get_url_data(serviceurl=serviceurl)

        return clients_stats

    @lru_cache(maxsize=32)
    def get_clients_reports_brief(self) -> list:

        # For multi server:
        # https://burp-ui.readthedocs.io/en/latest/api.html#get--api-clients-(server)-report
        # For single server:
        # https://burp-ui.readthedocs.io/en/latest/api.html#get--api-clients-report
        # GET /api/clients/report
        logging.debug("Url received: {}".format(self.apiurl))

        if self.IsMultiAgent:
            clients_report = self._get_clients_report_multi()
        else:
            serviceurl = self.apiurl + 'clients/report'
            clients_report = get_url_data(serviceurl=serviceurl)

        return clients_report

    @lru_cache(maxsize=32)
    def get_clients_running(self, server=None) -> list:
        """
        http://burp-ui.readthedocs.io/en/latest/api.html#get--api-clients-(server)-running
        :return: list of clients
        """

        if self.IsMultiAgent:
            serviceurl = '{}clients/{}/running'.format(self.apiurl, server)

        else:
            serviceurl = '{}clients/running'.format(self.apiurl)

        clients_running = get_url_data(serviceurl, ignore_empty=True)

        return clients_running

    @lru_cache(maxsize=32)
    def get_clients_reports(self):
        """
        Clients reports method
        :return list: [client_report_dict, ]
        example client:
        [ {'state': 'idle', 'human': '24 minutes ago', 'last': '2016-11-18T10:17:02+01:00', 'percent': 0,
            'server': 'Burp1', 'name': 'demo2',
        'backup_report': {'totsize': 8280079447,
            'efs': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'vssheader_enc': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'start': '2016-11-18T09:17:02+01:00', 'encrypted': False,
            'hardlink': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'windows': True,
            'vssheader': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'softlink': {'total': 102, 'deleted': 0, 'unchanged': 102, 'scanned': 102, 'changed': 0, 'new': 0},
            'files': {'total': 506894, 'deleted': 4, 'unchanged': 506882, 'scanned': 506894, 'changed': 10, 'new': 2},
            'end': '2016-11-18T09:19:49+01:00',
            'meta_enc': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'duration': 167, 'number': 2066,
            'meta': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'special': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'dir': {'total': 33094, 'deleted': 0, 'unchanged': 33087, 'scanned': 33094, 'changed': 0, 'new': 7},
            'vssfooter_enc': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'received': 1257441,
            'vssfooter': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
            'files_enc': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0}},
        'phase': None}]
        """

        # Get a list of clients to use
        clients_stats = self.get_clients_stats()

        logging.debug('clients_stats: {}'.format(clients_stats))

        # Create a new list to return
        clients_report = []
        
        for cli in range(len(clients_stats)):

            # Server, client is required to fetch report_stats
            server = clients_stats[cli].get('server', None)
            logging.debug('server: {}'.format(server))
            client = clients_stats[cli].get('name', None)

            # Omit getting stats for clients running a backup
            # burpui doesn't return data when client is running.
            server_running = self.get_clients_running(server)
            if client in server_running:
                continue

            # Create dict with all data of the client's dict
            client_report_dict = defaultdict(dict)
            client_report_dict.update(clients_stats[cli])
            # Create new list to use a list of numbers of backups only
            backups = []

            if not client:
                continue

            if clients_stats[cli].get('last', 'None') not in ['None', 'never']:

                # Get client_report_stats ;
                # It is a list of all backups stats
                cr_stats = self._get_client_report_stats(client, server=server)
                # and create a list of backups numbers only
                for n in range(len(cr_stats)):
                    backups.append(cr_stats[n].get('number'))

                if backups:
                    # Get the maximum number of backup to use
                    # The first backup could have date but not being reported with number, so no statistics.
                    # For that reason I add this if backups:
                    number = max(backups)

                    # Add the backup_report to the dict of the client
                    client_report_dict['backup_report'] = self._get_backup_report_stats(client, number, server=server)
                else:
                    client_report_dict['backup_report'] = self.empty_backup_report
            else:
                client_report_dict['backup_report'] = self.empty_backup_report

            if not client_report_dict['backup_report'].get('totsize'):
                client_report_dict['backup_report']['totsize'] = 0
            if not client_report_dict['backup_report'].get('received'):
                client_report_dict['backup_report']['received'] = 0
            if not client_report_dict['backup_report'].get('duration'):
                client_report_dict['backup_report']['duration'] = 0

            clients_report.append(client_report_dict)

        return clients_report



"""
Generates the default configuration for all sections.
"""
import configparser
import socket


def set_defaults():
    """
    Generates the default configuration for all sections.

    :return: configparser.ConfigParser object
    """

    config = configparser.ConfigParser(allow_no_value=True)

    inventory_status = {
        'spare': 'spare',
        'active': 'active',
        'inactive_in_burp': 'wrong not active',
        'inactive_not_in_burp': 'ignored inactive',
        'spare_in_burp': 'wrong spare in burp',
        'not_inventory_in_burp': 'not in inventory',
        'in_inventory_updated': 'ok',
        'spare_not_in_burp': 'ignored spare',
        'in_inventory_not_in_burp': 'absent',
        'in_many_servers': 'duplicated'
    }

    # Set dict of columns
    inventory_columns = {
        'client_name': 'device name',
        'status': 'status',
        'server': 'server',
        'sub_status': 'status (detailed)'
    }

    common = {
        'days_outdated': '31',
        'csv_delimiter': ';',
        # Options: https://docs.python.org/3.5/library/codecs.html#text-encodings
        # use mbcs for ansi on python prior 3.6
        'csv_encoding': 'utf-8',
        'excluded_clients': 'monitor,agent'
    }

    email_notification = {
        'email_to': 'root@localhost',
        'email_from': '{}@domain.com'.format(socket.gethostname()),
        'smtp_server': 'localhost',
        'smtp_port': '25',
        'smtp_login': '',
        'smtp_password': '',
        'smtp_mode': 'normal',
        'subject': 'burp report from {}'.format(socket.gethostname()),
        'foot_notes': 'A sample notes'
    }

    format_text = {
        'name_length': '15',
        'all_column_length': '11'
    }

    email_inventory = {
        'imap_host': 'localhost',
        'imap_password': 'password',
        'imap_user': 'username', # You can use 'domain\\user'
        'imap_folder': 'INBOX',
        'email_subject': 'inventory',
        'imap_search': 'TODAY', # TODAY will set today date in
        # format: "SENTON 23-Sep-2017 Subject \"inventory\"" (subject comes from email_subject key)
        # you could filter using the IMAP rules here (check
        # http://www.example-code.com/csharp/imap-search-critera.asp)
        'imap_port': '993',
        'attachment_save_directory': '/tmp',
        'attachment_filename': 'inventory.csv'
    }

    config['common'] = common
    config['inventory_columns'] = inventory_columns
    config['inventory_status'] = inventory_status
    config['email_notification'] = email_notification
    config['format_text'] = format_text
    config['email_inventory'] = email_inventory

    return config



#!python3
# https://gist.github.com/rdempsey/22afd43f8d777b78ef22
# https://stackoverflow.com/questions/3362600/how-to-send-email-attachments

import imaplib
import time
import os
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
# from email.utils import COMMASPACE, formatdate
from email.utils import formatdate
from email import encoders
from burp_reports.lib.imap import ImapReceive
from burp_reports.defaults.default_config import set_defaults

__inventory__ = os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..',
                                              'data', 'test_inventory.csv'))

defaults_settings = set_defaults()
defaults_settings['email_inventory']['imap_host'] = 'localhost'
defaults_settings['email_inventory']['imap_password'] = 'pwd1'
defaults_settings['email_inventory']['imap_user'] = 'test1'
defaults_settings['email_inventory']['imap_port'] = '3993'
defaults_settings['email_inventory']['imap_search'] = 'ALL'
config_test = dict(defaults_settings['email_inventory'])

defaults_settings['email_notification']['smtp_port'] = '3025'
defaults_settings['email_notification']['subject'] = 'inventory'
config_smtp = dict(defaults_settings['email_notification'])
# new_smtp_email = EmailNotifications(config=config_smtp, msg="test", attachments=[('utf-8', 'text', 'plain', 'burp_reports/data/test_inventory.csv', 'utf-8')])
# new_smtp_email.send_email()


def send_email():
    msg = MIMEMultipart()
    msg['From'] = "hello@itsme.com"
    msg['Subject'] = 'inventory'
    msg['Date'] = formatdate(localtime=True)
    files = [__inventory__]
    msg.attach(MIMEText("test message"))

    for file in files:
        part = MIMEBase('application', "octet-stream")
        part.set_payload(open(file, "rb").read())
        encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="{0}"'.format(
            os.path.basename(file)))
        msg.attach(part)

    conn = ImapReceive(config=config_test)
    conn_imap = conn._connect_imap()
    conn_imap.append('INBOX', '',
                     imaplib.Time2Internaldate(time.time()),
                     str(msg.as_string()).encode('utf-8'))

class TestImap:

    def test_connection(self):
        conn = ImapReceive(config=config_test)
        conn_test = conn._connect_imap()
        assert conn_test.check()[0] == 'OK'

    def test_attachment(self):
        send_email()
        imap_receive = ImapReceive(config=config_test)
        if os.path.isfile(imap_receive.save_file):
            os.remove(imap_receive.save_file)

        imap_receive.download_attachment()
        assert os.path.isfile("/tmp/inventory.csv")



#!python3

from .test_clients_dummy import test_dummy
from ..lib.configs import get_all_config
from ..reports.clients_reports import BurpReports
from ..lib.email import EmailNotifications
import pytest

class TestEmail:

    def test_compose(self):
        clients_dict = test_dummy()

        config = get_all_config()

        # Generate burp_reports object to use for reports.
        burp_reports = BurpReports(clients_dict,
                                   days_outdated=int(config['common']['days_outdated']),
                                   config=config)

        email = burp_reports._compose_email('outdated')

        assert isinstance(email, EmailNotifications)

    def test_compose_noconfig(self):
        clients_dict = test_dummy()

        # Generate burp_reports object to use for reports.
        burp_reports = BurpReports(clients_dict)

        with pytest.raises(SystemExit):
            burp_reports._compose_email('text')

    def test_send_email(self):
        clients_dict = test_dummy()
        config = get_all_config()
        config['email_notification']['smtp_port'] = '3025' # greenmail test smtp port

        # Generate burp_reports object to use for reports.
        burp_reports = BurpReports(clients_dict,
                                   days_outdated=int(config['common']['days_outdated']),
                                   config=config)

        send_email = burp_reports.email_outdated()

        assert send_email[0] or not send_email[0]



import pytest
from ..lib.urlget import get_url_data


def test_wrong_url():
    """
    Raise ValueError when return wrong value
    """

    with pytest.raises(ValueError):
        get_url_data('https://admin:admin@demo.burp-ui.org/api/demo1/c')


def test_wrong_url2():
    """
    Raise SystemExit
    """

    with pytest.raises(Exception):
        get_url_data('https://github.com/pablodav/cl')


def test_wrong_url_timeout():
    """
    Raise SystemExit
    """

    with pytest.raises(Exception):
        get_url_data('https://httpbin.org/delay/10', timeout=1)




def humanize_file_size(size):
    # It will be used on some prints of text for size representation.
    import math
    size = abs(size)
    if size == 0:
        return "0B"
    units = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
    p = math.floor(math.log(size, 2) / 10)
    return "%.3f%s" % (size / math.pow(1024, p), units[int(p)])



#!python3

from .test_clients_dummy import test_dummy
from ..lib.configs import get_all_config
from ..reports.clients_reports import BurpReports


class TestReports:

    def test_burpreports(self):
        clients_dict = test_dummy()

        config = get_all_config()

        # Generate burp_reports object to use for reports.
        burp_reports = BurpReports(clients_dict,
                                   days_outdated=int(config['common']['days_outdated']),
                                   config=config)

        return burp_reports

    def test_basic_txt(self):
        reports = self.test_burpreports()
        reports.print_basic_txt()

    def test_get_outdated(self):
        reports = self.test_burpreports()
        outdated = reports._get_outdated()

        assert isinstance(outdated, dict)

    def test_outdated_ping(self):
        reports = self.test_burpreports()
        reports.report_outdated(ping=True)

    def test_burpreports_noconfig(self):
        clients_dict = test_dummy()

        # Generate burp_reports object to use for reports.
        burp_reports = BurpReports(clients_dict,
                                   )

        return burp_reports



from datetime import datetime
from .humanize import humanize_file_size
from ..defaults.default_data_structure import default_client_backup_report
import os


class TxtReports:
    """
    Formats a dict of clients and prints to stdout or exports to file
    """

    def __init__(self, clients,
                 file=None,
                 detail=None,
                 additional_columns=None,
                 foot_notes='',
                 format_text=None):
        """
        :param clients: dict of clients formatted for burp reports.
        :param file: file to export the printed text
        :param detail: Use True when more detailed information is required
        :param additional_columns: Add more column in format: {'Column name': 'key'}
                Where 'key' is the key to search in, example: clients['key']
        :param foot_notes: str with more notes in the foot of the text
        :param format_text: dict with some settings for format text like: 
               { 'name_length': '15', 'all_column_length': '11' }
        """
        self.clients = clients
        self.file = file
        self.detail = detail
        self.additional_columns = additional_columns
        self.foot_notes = foot_notes
        self.empty_backup_report = default_client_backup_report()

        # set the format_text dict:
        default_format_text = {
            'name_length': '15',
            'all_column_length': '11'
        }
        
        self.format_text = format_text or default_format_text

    def format_client_text(self, client=None, header=None):
        """
        print only header once with client=None, header=True
        header to file:
        print_text(client=None, file=file, header=True)
        Recommended doc: https://pyformat.info/

        :param client: client_name to report
        :param header: True/None to report header formatted
        :return: client/header str depending on the option chosen
        """

        nlength = int(self.format_text.get('name_length')) # column length for name
        clength = int(self.format_text.get('all_column_length')) # column length for all columns except name
        headers_text = ''
        client_text = ''

        # List with dict {header: report_key to use from reports dict}
        client_details = [{'Date(local)': 'b_date'},
                          {'Time(local)': 'b_time'},
                          {'State': 'b_state'},
                          {'Phase': 'b_phase'}]

        empty_values = ('null', 'Null', 'None')

        if self.additional_columns:
            client_details.append(self.additional_columns)

            # if self.detail:
            # Extend more data to client_details list
            # Will be appended to both: header and data on report
            # client_details.append({'exclude': 'exclude'})

        # Format basic header
        if header:
            headers_text = str('\n burp report'.ljust(clength * 9) + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + '\n')

            # The clients name will be centered if there is detail, so the string starts different for client
            if self.detail:
                headers_text += ' {:>{}.{}} '.format('Name   ', nlength, nlength)
            else:
                headers_text += ' {:>{}.{}} '.format('Name   ', nlength * 2, nlength)

            # Look on each item on client_details list
            for n in range(len(client_details)):
                detail_dict = client_details[n]
                for k in detail_dict.keys():
                    # Append header from clients_details
                    headers_text += ' {:{}.{}} '.format(k, clength, clength)

            # Add more details from static method
            if self.detail:
                headers_text += self._txt_header_details(clength)

        # Format basic client information for the basic header
        if client:
            # Dict of client only data
            client_data = self.clients[client]

            # The clients name will be centered if there is detail, so the string
            # starts different for client
            if self.detail:
                client_text = ' {:>{}.{}} '.format(client, nlength, nlength)
            else:
                client_text = ' {:>{}.{}} '.format(client, nlength * 2, nlength)

            # Look on each item on client_details list
            for n in range(len(client_details)):
                detail_dict = client_details[n]
                for key, value in detail_dict.items():
                    # Add client data to the string client_text using client_details
                    # list to fetch keys
                    item_value = client_data.get(value, ' --- ')
                    # Check if item value is valid to be a string, if not add empty str value.
                    if not item_value or item_value in empty_values:
                        item_value = ' --- '
                    # Add the value of the item from client dictionary to client_text
                    # :Will put left , .Will truncate
                    client_text += ' {:{}.{}} '.format(item_value, clength, clength)

            # Add more details from static method
            if self.detail:
                client_text += self._txt_client_details(client_data=client_data, clength=clength)

        # Return formatted text
        if header:
            return headers_text
        if client:
            return client_text

    def _foot_notes(self, footer):
        """

        :param footer: (str) to add in foot notes
        :return: str formatted to be in foot_notes
        """
        footer_text = ''

        if self.detail:
            footer_text = '\n\n{}\n'.format(footer)

        if self.foot_notes:
            footer_text += '\n\n{}\n'.format(self.foot_notes)

        return footer_text

    @staticmethod
    def _txt_header_details(clength):
        """

        :param clength: number of spaces to use to justify and format columns
        :return:
        """
        # Additional details to add on headers and clients to the end of the strings
        headers_text = ' {:{}.{}} '.format('Duration', clength, clength)
        headers_text += ' {:{}.{}} '.format('Size', clength, clength)
        headers_text += ' {:{}.{}} '.format('Received', clength, clength)

        return headers_text

    def _txt_client_details(self, client_data, clength):
        """

        :param client_data: dict in burp reports format only for the client you want to get additional details
        :param clength: number of spaces to use to justify and format columns
        :return:
        """
        # Additional details to add on headers and clients to the end of the strings
        # Additional calculated added data
        # Look into client_data['backup_report']['duration']
        s = int(client_data.get('backup_report', self.empty_backup_report).get('duration', 0))
        duration = str('{:02}:{:02}:{:02}'.format(s // 3600, s % 3600 // 60, s % 60))
        # Look into client_data['backup_report']['totsize']
        totsize = int(client_data.get('backup_report', self.empty_backup_report).get('totsize', 0))

        # Look into client_data['backup_report']['received']
        received = int(client_data.get('backup_report', self.empty_backup_report).get('received', 0))

        # Additional calculated added data
        client_text = ' {:^{}} '.format(duration, clength, clength)
        client_text += ' {:<{}} '.format(humanize_file_size(totsize), clength)
        client_text += ' {:<{}} '.format(humanize_file_size(received), clength)

        return client_text

    def __print_text__(self, client=None, header=None, footer=None, print_text=True):
        """
        print only header once with client=None, header=True
        header to file:
        print_text(client=None, file=file, header=True)

        :param client:
        :param footer:
        :param header: print header
        :param print_text: to print or return
        :return: text line if no print_text
        """

        file_obj = None

        if self.file:
            if header:
                file_obj = open(self.file, 'w')
            else:
                file_obj = open(self.file, 'a')

        if header:
            header_text = self.format_client_text(client=None, header=True)
            if print_text:
                print(header_text, file=file_obj)
            else:
                return header_text

        if client:
            client_text = self.format_client_text(client, header=None)
            if print_text:
                print(client_text, file=file_obj)
            else:
                return client_text

        if footer:
            footer_text = self._foot_notes(footer=footer)
            if print_text:
                print(footer_text, file=file_obj)
            else:
                return footer_text

    def report_to_txt(self, print_text=True):
        """
        Method used to format the clients to text in stdout

        """

        total_taken = 0
        totsize = 0
        total_clients = 0
        text_body = ''
        foot_notes = ''

        if print_text:
            self.__print_text__(client=None, header=True)
        else:
            text_body += self.__print_text__(client=None, header=True, print_text=False)
            text_body += '\n'

        for client, v in sorted(self.clients.items()):
            client_data = self.clients[client]

            if print_text:
                self.__print_text__(client)
            else:
                text_body += self.__print_text__(client, print_text=False)
                text_body += '\n'

            if self.detail:
                total_taken += int(client_data.get('backup_report', {}).get('duration', 0))
                totsize += int((client_data.get('backup_report', {}).get('totsize', 0)))
                total_clients += 1

        if self.detail:
            foot_notes = '\n{:>30}  {:02}:{:02}:{:02}'.format('total duration backups taken:',
                                                              total_taken // 3600,
                                                              total_taken % 3600 // 60,
                                                              total_taken % 60)

            foot_notes += "\n{:>30}  {}".format('total size in backup:',
                                                humanize_file_size(totsize))

            foot_notes += '\n{:>30}  {}'.format('Total clients:',
                                                total_clients)

        if foot_notes:

            if print_text:
                self.__print_text__(client=None, footer=foot_notes)
            else:
                text_body += self.__print_text__(client=None, footer=foot_notes, print_text=False)
                text_body += '\n'

        if self.file:
            if os.path.isfile(self.file):
                print('exported to', self.file)

        if not print_text:
            return text_body



from ..__main__ import parse_args
import pytest


class TestParser:

    def test_noargs(self):
        """
        User passes no args, should fail with SystemExit
        """
        with pytest.raises(SystemExit):
            options = parse_args([])

    def test_ui_dummy(self):
        """
        -ui dummy must be tested

        :return argparse dict with options and burpui_apiurl = dummy
        """
        options = parse_args(['-ui', 'dummy'])
        assert options.burpui_apiurl == 'dummy'




from ..lib.check_readme import check_readme
import os
import pytest


__readme__ = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..', 'README.rst')


def test_check_readme():
    check_readme(__readme__)


def test_check_readme_error():
    __readmef__ = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'Readme.rst')
    with pytest.raises(SystemExit):
        check_readme(__readmef__)



#! python3
import configparser
from ..defaults.default_config import set_defaults
import os


def parse_config2(filename=None):
    """
    https://docs.python.org/3.5/library/configparser.html

    :param filename: filename to parse config
    :return: config_parse result
    """

    config = configparser.ConfigParser(allow_no_value=True)

    if filename:
        # ConfigParser does not create a file if it doesn't exist, so I will create an empty one.
        if not os.path.isfile(filename):
            with open(filename, 'w', encoding='utf-8') as f:
                print('', file=f)

        config.read_file(open(filename))

    return config


def get_all_config(filename=None):
    """
    Set default configuration for burp_reports
    Config with defaults settings if no file will be passed
    Also with defaults sections and defaults keys for missing options in config

    :param filename: options config file to read
    :return: config with default config for missing sections
    """

    config = parse_config2(filename)
    default_config = set_defaults()

    # Verify each section in default_config
    for s in range(len(default_config.sections())):
        section = default_config.sections()[s]

        # Add the missing section to the config obtained
        if not config.has_section(section):
            config.add_section(section)

        # Add missing keys to config obtained
        for key in default_config[section]:
            if not config.has_option(section, key):
                config[section][key] = default_config[section][key]

    return config



# -*- coding: utf8 -*-
import logging
from .burpui_api_translate import TranslateBurpuiAPI
from ..backends.burpui_api import Clients


class BUIClients:
    """"
    Get data from burp ui clients
    """

    def __init__(self, burpui_apiurl):
        """

        """

        # Define clients from Interface
        self.clientsobj = Clients(apiurl=burpui_apiurl)

    def translate_clients_stats(self, detail=None):
        """

        :return: clients translated
        """

        # Get the list of clients from the Interface
        if detail:
            clients = self.clientsobj.get_clients_reports()
        else:
            clients = self.clientsobj.get_clients_stats()

        logging.debug('clients: {}'.format(clients))
        # Set clients list from api interface to TranslateBurpuiAPI object
        clients_list_api = TranslateBurpuiAPI(clients=clients)
        # Translate with method translate_clients()
        clients_reports = clients_list_api.translate_clients()
        return clients_reports



# -*- coding: utf8 -*-


class Clients:
    """
    Get clients information.

    """

    def __init__(self, burp_version=1, conf=None):
        """

        :param burp_version: version of burp backend to work with 1/2
        :param conf: burp_ui configuration to use.
        """

        self.version = burp_version
        self.buiconf = conf

    @staticmethod
    def get_client(client='monitor'):
        """

        :param client: Name of the client to retrieve data.
        :return: [{'received': 326753806, 'end': 1466714070, 'encrypted': False, 'number': 1, 'deletable': True,
        'date': 1466713986, 'size': 572911431}]
        """

        client_data = [{'received': 326753806, 'end': 1466714070, 'encrypted': False, 'number': 1, 'deletable': True,
                        'date': 1466713986, 'size': 572911431}]
        return client_data

    @staticmethod
    def get_clients():
        """
        #  server.get_all_clients()
        #
        :return: [{'state': u'idle', 'last': 1466703186, 'name': u'monitor'}]
        """

        all_clients = [{'state': u'idle', 'last': 1466703186, 'name': u'monitor'}]

        return all_clients



# encoding: utf-8
# https://docs.python.org/3.5/library/imaplib.html#imap4-example
# https://gist.github.com/jasonrdsouza/1674794

import imaplib
import email
import os
import arrow


class ImapReceive:
    def __init__(self, config):
        """
        :param msg: (str)
        :param config: (dict) dict with configurations

        {   'imap_host': 'hostname',
            'imap_password': 'password',
            'imap_user': 'username', # You can use 'domain\\user'
            'imap_folder': 'INBOX',
            'email_subject': 'inventory',
            'imap_search': 'TODAY', # TODAY will set today date in
            format: "SENTON 23-Sep-2017 Subject \"inventory\""
            # you could filter using the IMAP rules here (check
            # http://www.example-code.com/csharp/imap-search-critera.asp)
            'imap_port': '993',
            'attachment_save_directory': '/tmp',
            'attachment_filename': 'inventory.csv'
        }

        """

        self.host = config['imap_host']
        self.user = config['imap_user']
        self.password = config['imap_password']
        self.imap_folder = config['imap_folder'] or 'INBOX'
        self.email_subject = config['email_subject'] or 'inventory'
        self.imap_search = config['imap_search'] or 'TODAY'
        self.imap_port = config['imap_port'] or '993'
        self.save_directory = config['attachment_save_directory'] or '.'
        self.filename = config['attachment_filename'] or 'inventory.csv'
        self.save_file = os.path.join(self.save_directory, self.filename)

        if config['imap_search'] == 'TODAY':
            today_string = arrow.now().format('DD-MMM-YYYY')
            self.imap_search = "SENTON {} Subject \"{}\"".format(today_string, self.email_subject)

    def _connect_imap(self):

        conn = imaplib.IMAP4_SSL(self.host, self.imap_port)
        conn.login(self.user, self.password)
        conn.select(self.imap_folder)

        return conn

    def download_attachment(self):

        conn = self._connect_imap()

        # results,data = conn.search(None,'ALL')
        # resp, data = conn.search(None, "ALL") # you could filter using the IMAP rules here
        # (check http://www.example-code.com/csharp/imap-search-critera.asp)
        resp, data = conn.search(None, self.imap_search)

        msg_ids = data[0]
        msg_id_list = msg_ids.split()

        for emailid in msg_id_list:
            # fetching the mail, "`(RFC822)`" means "get the whole stuff", but you can ask for
            # headers only, etc
            resp, data = conn.fetch(emailid, "(RFC822)")
            email_body = data[0][1] # getting the mail content
            # mail = email.message_from_string(email_body) # parsing the mail content to get a
            # mail object
            mail = email.message_from_bytes(email_body)  # parsing the mail content to
            # get a mail object python3
            # https://docs.python.org/3.5/library/email.html

            #Check if any attachments at all
            if mail.get_content_maintype() != 'multipart':
                continue

            #    print ("["+mail["From"]+"] :" + mail["Subject"])

            # we use walk to create a generator so we can iterate on the parts and forget
            # about the recursive headach
            for part in mail.walk():
                # multipart are just containers, so we skip them
                if part.get_content_maintype() == 'multipart':
                    continue

                # is this part an attachment ?
                if part.get('Content-Disposition') is None:
                    continue

                #filename = part.get_filename()
                #filename = mail["From"] + "_hw1answer"

                att_path = self.save_file

                #Check if its already there
                #if not os.path.isfile(att_path):

                # finally write the stuff
                with open(att_path, 'wb') as file_open:
                    file_open.write(part.get_payload(decode=True))
                # Save the attachment only for one file
                break

        conn.close()



#!python3

from ..interfaces.burpui_api_interface import BUIClients

burpui_demo='https://admin:admin@demo.burp-ui.org/api/'
burpui_standalone='http://admin:admin@172.17.0.2:8080/api/'

def test_bui_demo():
    clients_obj = BUIClients(burpui_apiurl=burpui_demo)
    clients_dict = clients_obj.translate_clients_stats()

    assert isinstance(clients_dict, dict)

    return clients_dict

def test_bui_demo_detail():
    clients_obj = BUIClients(burpui_apiurl=burpui_demo)
    clients_dict = clients_obj.translate_clients_stats(detail=True)

    assert isinstance(clients_dict, dict)

    return clients_dict



from ..lib.is_up import is_up
from ..lib.is_up import outdated_pings


class TestIsUP(object):

    def test_isup_ok(self):
        assert is_up('localhost')

    def test_isup_fail(self):
        assert not is_up('nohost')

    def test_isup_feedback_ok(self):
        assert is_up('localhost', give_feedback=True)

    def test_isup_feedback_fail(self):
        assert not is_up('nohost', give_feedback=True)

    def test_outdated_pings(self):
        clients = {'localhost': {}}
        result = outdated_pings(clients)
        assert isinstance(result['localhost']['comments'], str)




#!python3

from ..dummy.burpui_dummy_api import Clients


class TestDummy:

    def test_dummy_api(self):
        clients_obj = Clients
        clients_dict = clients_obj.get_clients()

        assert clients_dict[0]['name'] == 'monitor'



#!python3
# 

# Default dict for 

def default_client_backup_report():
    empty_backup_report = dict()
    empty_backup_report = {'totsize': 0,
                       'efs': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'vssheader_enc': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'start': '0', 'encrypted': False,
                       'hardlink': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'windows': 'Unknown',
                       'vssheader': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                        'softlink': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'files': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'end': '0',
                       'meta_enc': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'duration': 0, 'number': 0,
                       'meta': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'special': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'dir': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'vssfooter_enc': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'received': 0,
                       'vssfooter': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0},
                       'files_enc': {'total': 0, 'deleted': 0, 'unchanged': 0, 'scanned': 0, 'changed': 0, 'new': 0}
                       }
    
    return empty_backup_report



import tempfile
import os
import getpass


def temp_file(file='temporal'):
    """
    return: str with tempfilename
    """
    # Append uid to end of filename
    file += '_{}'.format(getpass.getuser())
    # Simplified and reutilized core funtionally from python
    cache_path = os.path.join(tempfile.gettempdir(), file)

    return cache_path



import rstcheck


def check_readme(file='README.rst'):
    """
    Checks readme rst file, to ensure it will upload to pypi and be formatted correctly.
    :param file:
    :return:
    """
    # Get the long description from the relevant file
    with open(file, encoding='utf-8') as f:
        readme_content = f.read()

    errors = list(rstcheck.check(readme_content))
    if errors:
        msg = 'There are errors in {}, errors \n {}'.format(file, str(errors[0]))
        raise SystemExit(msg)
    else:
        msg = 'No errors in {}'.format(file)
        print(msg)




import os
import platform
from functools import lru_cache
from concurrent.futures import ThreadPoolExecutor as Executor


@lru_cache(None)
def is_up(hostname, give_feedback=False):
    """

    :param hostname: hostname/ip to ping
    :param give_feedback: just print up or down to stdout
    :return: True/False
    """

    if platform.system() == "Windows":
        response = os.system("ping " + hostname + " -n 1")
    else:
        response = os.system("ping -c 1 " + hostname)

    is_up_bool = False

    # If ping responds
    if response == 0:
        if give_feedback:
            print(hostname, 'is up!')
        is_up_bool = True
    else:
        if give_feedback:
            print(hostname, 'is down!')

    return is_up_bool


# Functions to add parallel ping:
def check_isup(k, return_client=None):
    """
    Checks ping and returns status
    Used with concurrent decorator for parallel checks

    :param k: name to ping
    :param return_client: to change return format as '{k: {'comments': comments}}'
    :return(str): ping ok / -
    """
    if is_up(k):
        comments = 'ping ok'
    else:
        comments = ' - '

    if return_client:
        comments = {k: {'comments': comments}}

    return comments


def outdated_pings(outdated_clients: dict):
    """
    Appends comments to clients if pings or not.
    Used with synchronized decorator for parallel checks (required def with concurrent decorator)

    :param outdated_clients: dict formatted with clients as keys
    :return: dict with appended comments
    """
    outdated = outdated_clients

    with Executor(max_workers=10) as exe:
        # jobs are executed in parallel with exe.submit
        # params passed to exe.submit are: (function, args, args)
        # in this case we pass client, True. True is to get a return as dict 
        # with key as client and new comments
        # we use one are from a list appending a for in the list
        jobs = [exe.submit(check_isup, k, True) for k in outdated.keys()]
        # results are generated from jobs
        results = [job.result() for job in jobs]

    for i, values1 in enumerate(results):
        data_client = list(values1.items())
        client = data_client[0][0]
        comment = data_client[0][1].get('comments')
        # Append ping information to outdated_clients
        # as we have same format in results as the original dict, we just update the values.
        # dict returned by results is {k: {'comments': comments}}
        outdated[client]['comments'] = comment

    return outdated



from ..lib.txt import TxtReports
from .test_clients_dummy import test_dummy


class TestTxt:
    
    def test_detail_txt(self):
        clients = test_dummy()
        # Create the object to export the report
        clients_reports = TxtReports(clients,
                                     detail=True,
                                     foot_notes="This is a test foot txt"
                                     )

        clients_reports.report_to_txt(print_text=True)

    def test_detail_txt_file(self):
        clients = test_dummy()
        # Create the object to export the report
        clients_reports = TxtReports(clients,
                                     detail=True,
                                     foot_notes="This is a test foot txt",
                                     file='test1.txt'
                                     )

        clients_reports.report_to_txt()

    def test_detail_txt_debug(self):
        clients = test_dummy()
        import logging
        logging.basicConfig(level=logging.DEBUG)
        # Create the object to export the report
        clients_reports = TxtReports(clients,
                                     detail=True
                                     )

        clients_reports.report_to_txt(print_text=True)



# encoding: utf-8
# http://www.magiksys.net/pyzmail/

import pyzmail


# import email
# import time


class EmailNotifications:
    def __init__(self, config, msg, attachments=None):
        """
        :param msg: (str)
        :param config: (dict) dict with configurations

        {   'email_to': 'root@localhost',
            'smtp_password': '',
            'email_from': 'server@domain.com',
            'smtp_server': 'localhost',
            'smtp_login': '',
            'smtp_mode': 'normal',
            'smtp_port': '25',
        }
        :param attachments: [] files to attach attachments (list) - the list of attachments to include into the mail,
        in the form [(data, maintype, subtype, filename, charset), ..] where :

        data : is the raw data, or a charset encoded string for 'text' content.
        maintype : is a MIME main type like : 'text', 'image', 'application' ....
        subtype : is a MIME sub type of the above maintype for example : 'plain', 'png', 'msword' for respectively
        'text/plain', 'image/png', 'application/msword'.
        filename this is the filename of the attachment, it must be a 'us-ascii' string or a tuple of the form
        (encoding, language, encoded_filename) following the RFC2231 requirement, for example ('iso-8859-1',
        'fr', u'r\xe9pertoir.png'.encode('iso-8859-1'))
        charset : if maintype is 'text', then data must be encoded using this charset. It can be None for non 'text'
        content.
        """

        self.fromaddr = config['email_from']
        self.toaddr = config['email_to'].split(',')
        self.smtp_host = config['smtp_server']
        self.smtp_port = int(config.get('smtp_port', 25))
        self.smtp_mode = config.get('smtp_mode', None)
        self.smtp_login = config.get('smtp_login', None)
        self.smtp_password = config.get('smtp_password', None)

        self.sender = self.fromaddr
        self.recipients = self.toaddr
        self.subject = config.get('subject', 'the subject')

        self.attachments = attachments
        self.msg = msg

    def send_email(self):
        """

        :return: tuple, True/False, details
        """

        payload, mail_from, rcpt_to, msg_id = self.compose_email()

        # Send the email:
        ret = pyzmail.send_mail(payload, mail_from, rcpt_to,
                                smtp_host=self.smtp_host,
                                smtp_port=self.smtp_port,
                                smtp_mode=self.smtp_mode,
                                smtp_login=self.smtp_login,
                                smtp_password=self.smtp_password)

        if isinstance(ret, dict):
            if ret:
                return False, 'failed recipients:', ', '.join(ret.keys())
            else:
                return True, 'success'
        else:
            return False, 'error:', ret

    def compose_email(self):
        # Define variables to use

        text_content = self.msg
        preferred_encoding = 'iso-8859-1'
        text_encoding = 'iso-8859-1'
        # date = email.utils.formatdate(time.time(), localtime=True)
        # Compose the email in payload, mail_from, rcpt_to, msg_id

        payload, mail_from, rcpt_to, msg_id = pyzmail.compose_mail(
            (self.sender, self.sender),
            self.recipients,
            self.subject,
            preferred_encoding,
            (text_content, text_encoding),
            html=None,
            attachments=self.attachments)

        return payload, mail_from, rcpt_to, msg_id



# -*- coding: utf8 -*-
import json
import time
from datetime import timedelta

import requests
import requests_cache
from urllib3.util import parse_url

from .files import temp_file

cache_file = 'burp_reports_cache'
expire_after = timedelta(minutes=30)
cache_path = temp_file(cache_file)

requests_cache.install_cache(cache_path, backend='sqlite', expire_after=expire_after)


def get_url_data(serviceurl: 'url to retrieve data',
                 params: "python requests params in url" = None,
                 ignore_empty: "returned [] value will be ignored" = False,
                 timeout: "how much time to wait for a response" = 60,
                 check_multi: "check if response has data or not" = False):

    """

    :param serviceurl: url to retrieve data
    :param params: http://docs.python-requests.org/en/master/user/quickstart/#passing-parameters-in-urls
    :param ignore_empty: returned [] value will be ignored, so it will allow return [] from url
    :param timeout: how much time to wait for a response
    :return: json url_data
    """

    # Get data from the url
    # Support https without verification of certificate
    # req = requests.get(serviceurl, verify=False, params=params)
    retry_times = 0
    max_retry = 4
    purl = parse_url(serviceurl)
    message = ''
    req = []

    if purl.auth:
        username = purl.auth.split(':')[0]
        password = purl.auth.split(':')[1]
    else:
        username = None
        password = None
    # Add url like http://host
    burl = '{}://{}'.format(purl.scheme, purl.host)
    if purl.port:
        # Add port like: http://host:8080
        burl += ':{}'.format(purl.port)
    if purl.request_uri:
        # Add path and query like: http://host:8080/path/uri?query
        burl += '{}'.format(purl.request_uri)

    while retry_times < max_retry:

        message = ''
        retry_times += 1

        try:
            req = requests.get(burl, verify=False, params=params, timeout=timeout, auth=(username, password))
            if check_multi:
                if not req.json():
                    return []

            req.json()

            if req.json():

                if isinstance(req.json(), dict):
                    message = req.json().get('message', '')

                elif isinstance(req.json(), list):

                    if isinstance(req.json()[0], dict):
                        message = req.json()[0].get('message', '')

                if message in ['timed out']:
                    # next try
                    requests_cache.clear()
                    time.sleep(2)
                    continue

                # Don't try again
                break

            elif ignore_empty:
                continue

            elif req.from_cache:
                # Clear cache to retry again
                # Added in urlget module test if it's [] retry n times due to issue:
                # https://git.ziirish.me/ziirish/burp-ui/issues/148
                requests_cache.clear()
                time.sleep(2)
                # next try
                continue

            else:
                # Raise a custom exception
                raise ValueError('No data from response')

        except requests.exceptions.RequestException as e:

            time.sleep(2)

            print('request failed to {} \n retry Nº: {}'.format(burl, retry_times))

            if retry_times >= max_retry:
                raise e

        except ValueError as e:

            print('request failed to get {}'.format(burl))
            raise e

        except Exception as e:

            e('request failed to {} \n with exception'.format(burl))

    if message == 'timed out':
        raise TimeoutError('request timed out')

    data = req.json()

    return data



# Help used to build it:
# https://chriswarrick.com/blog/2014/09/15/python-apps-the-right-way-entry_points-and-scripts/
# https://docs.python.org/3/distributing/index.html#reading-the-guide

import sys
import os
import argparse
import logging
from .lib.configs import parse_config2
from .lib.configs import get_all_config
from .reports.clients_reports import BurpReports
from collections import defaultdict
from . import __version__


def parse_args(args):
    """
    Information extracted from: https://mkaz.com/2014/07/26/python-argparse-cookbook/
    :return:
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-c', '--reports_conf', dest='reports_conf',
                        const=os.path.join(os.sep, 'etc', 'burp', 'burp-reports.conf'),
                        default=None,
                        nargs='?', help='burp-reports.conf configuration file')

    parser.add_argument('-ui', '--burpui_apiurl', dest='burpui_apiurl', nargs='?', default=None, const=None,
                        help='full url to burpui api, like http://user:pass@server:port/api/ \n'
                             'It is required on cli line or in config file\n\n')

    # Adding report choices

    choices_helper = {"print": "Print txt clients list only \n\n",
                      "outdated": "will print list of outdated clients \n",
                      "inventory":
                          'requires -i and -o, check input inventory and generates a comparison\n'
                          '    Input csv headers required: device name; status; Status (detailed) \n'
                          '       Example line:            demo1; active; \n'
                          '                                demo2; active; spare \n'
                          '     -i email_inventory will download email from imap account set in config',
                      "email_outdated": "Will send email to configured recipients on config file\n"}

    parser.add_argument("--report", '-r', dest='report', nargs='?', default='print', const='print',
                        choices=choices_helper,
                        help='\n'.join("{}: {}".format(key, value) for key, value in sorted(choices_helper.items())))

    parser.add_argument('--debug', dest='debug', nargs='?', default=None, const=True,
                        help='Activate for debugging purposes')

    parser.add_argument('--version', dest='version', nargs='?', default=None, const=True,
                        help='Print version and exit')

    parser.add_argument('--detail', dest='detail', nargs='?', default=None, const=True,
                        help='Adds more details to reports')

    parser.add_argument('--ping', dest='ping', nargs='?', default=None, const=True,
                        help='Adds ping check to outdated report only')

    parser.add_argument('--write_config', dest='write_config', default=None, action='store_true',
                        help="Write configuration with default values, useful to get a config file to modify")

    parser.add_argument('-i', nargs='?', default=None, help='Input csv file to use on --report inventory \n'
                                                            '(also can be an url to download it)\n')
    parser.add_argument('-o', nargs='?', default=None, help='Output csv file to use on --report inventory')

    if not args:
        raise SystemExit(parser.print_help())

    return parser.parse_args(args)


def bui_api_clients_stats(burpui_apiurl, detail=None):
    """

    :param burpui_apiurl: string to burpui_apiurl, full url http://user:pass@server:port/api/
    :param detail: Adds more detailed info, backup_report nested dict with some info like: duration, received, totsize.
    :return: dict with clients stats
    """
    from .interfaces.burpui_api_interface import BUIClients
    bui_clients = BUIClients(burpui_apiurl=burpui_apiurl)

    clients_dict = bui_clients.translate_clients_stats(detail=detail)

    return clients_dict


def bui_dummy_clients_stats():
    # Get data from dummy module (for testing the functions only)
    from .dummy.burpui_api_translate_dummy import BUIClients
    clients_dict = BUIClients()
    clients_dict = clients_dict.translate_clients_stats()
    return clients_dict


def get_main_conf(options):
    """

    :param options: options from argparser
    :return: dict with options from config file or defaults
    """
    _options = defaultdict(dict)

    config_options = parse_config2(options.reports_conf)

    # Use general section for general options:
    if config_options.has_section('common'):
        general_config = config_options['common']
    else:
        general_config = {}

    # Creating default values for our config:
    _options['burpui_apiurl'] = general_config.get('burpui_apiurl', None)

    # burpui_apiurl to defined on command line options
    # Override the burpui_apiurl defined before
    if options.burpui_apiurl:
        _options['burpui_apiurl'] = options.burpui_apiurl

    return _options


def cli_execution(options):
    """
    Manage command line options for cli usage

    :param options:
    :return:
    """
    clients_dict = {}

    # Print version and exit with --version option
    if options.version:
        raise SystemExit('{}'.format(__version__))

    # Configs that can be overwritten by command line options
    config_options = get_main_conf(options)
    debug = options.debug

    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    burpui_apiurl = config_options.get('burpui_apiurl')
    # Config with defaults settings if no file will be passed
    # Also with defaults sections and defaults keys for missing options in config
    all_config = get_all_config(options.reports_conf)
    if options.write_config:
        if not options.reports_conf:
            raise SystemExit('--write_config requires -c configfile.conf')

        with open(options.reports_conf, 'w', encoding='utf-8') as f:
            all_config.write(f, space_around_delimiters=True)
    days_outdated = all_config.getint('common', 'days_outdated')
    # If there is an option to for burpui_apiurl, get clients from that apiurl
    if burpui_apiurl:
        if burpui_apiurl.lower() == 'dummy':
            clients_dict = bui_dummy_clients_stats()
        else:
            # Get clients stats from burpui_api_interface
            clients_dict = bui_api_clients_stats(burpui_apiurl, detail=options.detail)
    else:
        raise SystemExit('burpui_apiurl is required in cmd or in config common section')
    
    # Generate burp_reports object to use for reports.
    reports = BurpReports(clients_dict,
                          days_outdated=days_outdated,
                          detail=options.detail,
                          config=all_config)

    # Add some report option to use, use clients_dict already set
    if options.report == 'print':
        reports.print_basic_txt()

    elif options.report in ['outdated', 'o']:
        reports.report_outdated(export_txt=True,
                                ping=options.ping)

    elif options.report == 'inventory':
        reports.save_compared_inventory(options.i,
                                        options.o)

    elif options.report == 'email_outdated':
        reports.email_outdated()


def main():
    """
    Main function
    """
    options = parse_args(sys.argv[1:])

    cli_execution(options)


if __name__ == "__main__":
    main()



# -*- coding: utf8 -*-
import arrow
from collections import defaultdict
from functools import lru_cache


class TranslateBurpuiAPI:
    """
    Translate data from burpui_api to burp_reports
    """

    def __init__(self, clients):
        """
        Initialize vars
        :param: clients list of clients coming from burp_api
        """
        self.clients = clients

    def translate_clients_function(self, data_t):
        """

        :param data_t: list of clients dicts to use for translation
        :return: d_clients dictionary of clients translated to use in burp_reports
        example from clients stats:

        {'client_name':
            { 'b_last'    : 'YYYY-MM-DDTHH:mm:ssZZ',
              'b_state'    : 'working/current',
              'b_phase' : 'phase1/phase2',
              'b_date' : 'date(local)',
              'b_time' : 'time(local)',
              'server': 'server in multi-agent mode'
            }
        }
        """
        d_clients = defaultdict(dict)

        for cli in range(len(self.clients)):
            # Get the client dict data from list of clients
            client_data = self.clients[cli]
            client_name = client_data.get(data_t['client_name'], None)

            # Go to next item in client_data if there is not 'client_name'
            if not client_name:
                continue

            # Translate and define variables:
            # Define a dict with data to clients
            for k, v in data_t.items():

                if k is not 'client_name':
                    # similar and simplified to: d_clients.setdefault(client_name, {})['b_phase'] = b_phase
                    d_clients[client_name][k] = client_data.get(v, None)

                # Add b_date and b_time from b_last information
                # if k is b_last and b_last has data and is not never, add the keys.
                if k is 'b_last':
                    last_backup = client_data.get(v, None)
                    if last_backup and last_backup not in 'never':
                        if last_backup == 'now':
                            date_and_time = arrow.get()
                            d_clients[client_name]['b_last'] = date_and_time.isoformat(sep='T')
                        else:
                            date_and_time = arrow.get(last_backup)
                        # Convert date_and_time to local time
                        date_and_time = date_and_time.to('local')
                        d_clients[client_name]['b_date'] = date_and_time.format('YYYY-MM-DD')
                        d_clients[client_name]['b_time'] = date_and_time.format('HH:mm:ss')

            # Add server list if not exists
            if not d_clients[client_name].get('server', None):
                d_clients[client_name]['server'] = []
            # Append servers if exist
            if client_data.get('server', None):
                d_clients[client_name]['server'].append(client_data['server'])

        # Return dictionary of clients expected to use in burp_reports
        return d_clients

    def translate_clients(self):
        """

        :return:
        {'client_name':
            { 'b_last'    : '2016-06-23T14:33:06-03:00',
              'b_state'    : 'working/current',
              'b_phase' : 'phase1/phase2'
              'b_date' : 'local time date'
              'b_time' : 'local time'
              'backup_report' : 'dict with backup report, data like duration, totsize, received'
            }
        }
        """

        # Dictionary to use for translation
        data_t = {
            "b_phase": 'phase',
            "b_state": "state",
            "b_last": "last",
            "client_name": "name"}

        # Check if key backup_report exists in clients
        if 'backup_report' in self.clients[0]:
            # Add to dictionary of client
            data_t['backup_report'] = 'backup_report'

        reports_clients = self.translate_clients_function(data_t=data_t)
        return reports_clients



#!python3

from .test_clients_dummy import test_dummy
from ..lib.configs import get_all_config
from ..reports.clients_reports import BurpReports
import os
from ..lib.files import temp_file
from invpy_libs import csv_as_dict

# Read version from VERSION file
__inventory__ = os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..',
                                              'data', 'test_inventory.csv'))

__output__ = temp_file('inventory_test.csv')


class TestInventory:
    def test_file(self):
        assert os.path.isfile(__inventory__)

    def test_inventory(self):
        clients_dict = test_dummy()

        config = get_all_config()

        # Generate burp_reports object to use for reports.
        reports = BurpReports(clients_dict,
                              days_outdated=int(config['common']['days_outdated']),
                              config=config)

        reports.save_compared_inventory(__inventory__, __output__)
        compared_clients = csv_as_dict(__output__, config['inventory_columns']['client_name'],
                                       delimiter=config['common']['csv_delimiter'])

        return compared_clients

    def test_inventory_result(self):
        report = self.test_inventory()
        config = get_all_config()

        status_dict = dict(config['inventory_status'])
        status = config['inventory_columns']['status']

        assert report['cli10'][status] == status_dict['in_inventory_not_in_burp']
        assert report['cli10'][status] not in status_dict['in_many_servers']
        assert report['cli20'][status] == status_dict['spare_not_in_burp']
        assert report['cli30'][status] == status_dict['in_many_servers']
        assert report['cli40'][status] == status_dict['inactive_not_in_burp']
        assert report['client_001'][status] == status_dict['in_inventory_updated'] or 'outdated'
        assert report['client_002'][status] == status_dict['spare_in_burp']
        assert report['client_003'][status] == status_dict['inactive_in_burp']



#! python3
# pylint: disable=invalid-name
"""
 Setup file to package burp_reports
 Help from: http://www.scotttorborg.com/python-packaging/minimal.html
 https://docs.python.org/3/distutils/commandref.html#sdist-cmd
 https://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
 https://docs.python.org/3.4/tutorial/modules.html
 Install it with python setup.py install
 Or use: python setup.py develop (changes to the source files will be
 immediately available)
 https://pypi.python.org/pypi?%3Aaction=list_classifiers
"""
import os
from os import path
from setuptools import setup, find_packages
import rstcheck

here_path = path.abspath(path.dirname(__file__))

with open(os.path.join(here_path, 'requirements.txt')) as f:
    requires = [x.strip() for x in f if x.strip()]


def check_readme(file='README.rst'):
    """
    Checks readme rst file, to ensure it will upload to pypi and be formatted
    correctly.
    :param file:
    :return:
    """
    # Get the long description from the relevant file
    with open(file, encoding='utf-8') as f_object:
        readme_content = f_object.read()

    errors = list(rstcheck.check(readme_content))
    if errors:
        msg = 'There_path are errors in {}, errors \n {}'.format(file,
                                                            errors[0].message)
        raise SystemExit(msg)
    else:
        msg = 'No errors in {}'.format(file)
        print(msg)

readme_path = path.join(here_path, 'README.rst')

# Get the version from VERSION file
mypackage_root_dir = 'burp_reports'
with open(os.path.join(mypackage_root_dir, 'VERSION')) as version_file:
    version = version_file.read().strip()

# Get the long description from the relevant file
with open(readme_path, encoding='utf-8') as f:
    long_description = f.read()

check_readme(readme_path)

# Define setuptools specifications
setup(name='burp_reports',
      version=version,
      description='Burp reports package',
      long_description=long_description,  # this is the file README.rst
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3 :: Only',
          'Topic :: System :: Archiving :: Backup',
      ],
      url='https://github.com/pablodav/burp_server_reports',
      author='Pablo Estigarribia',
      author_email='pablodav@gmail.com',
      license='BSD3',
      packages=find_packages(),
      include_package_data=True,
      package_data={
          'data': 'burp_reports/data/*',
      },
      data_files=[('VERSION', ['burp_reports/VERSION'])],
      entry_points={
          'console_scripts': [
              'burp-reports = burp_reports.__main__:main'
          ]
      },
      install_requires=requires,
      tests_require=['pytest',
                     'pytest-cov'],
      zip_safe=False)



# -*- coding: utf8 -*-

from ..interfaces.burpui_api_translate import TranslateBurpuiAPI
from ..dummy.burpui_dummy_api import Clients


class BUIClients:
    """"
    Get data from burp ui clients
    """

    def __init__(self):
        """

        """
        # Define clients from Interface
        self.clientsobj = Clients
        # Get the list of clients from the Interface
        self.clients = self.clientsobj.get_clients()

    def translate_clients_stats(self):
        """

        :return: clients translated
        """

        # Set clients list from api interface to TranslateBurpuiAPI object
        clients_list_api = TranslateBurpuiAPI(clients=self.clients)
        # Translate with method translate_clients()
        clients_reports = clients_list_api.translate_clients()

        return clients_reports




# -*- coding: utf8 -*-
from ..defaults.default_data_structure import default_client_backup_report
import arrow
import random


class Clients:
    """
    Get clients from burpui api url.

    """

    def __init__(self):
        """

        """

    @staticmethod
    def get_clients():
        """
        #  server.get_all_clients()
        #
        :return: [{
        "phase": null,
        "percent": 0,
        "state": "idle",
        "last": "2016-06-23T14:33:06-03:00",
        "name": "monitor"}]
        """
        actual_time = arrow.get()

        default_backup_report = default_client_backup_report()

        clients_stats = [{
            "phase": 'null',
            "percent": 0,
            "state": "idle",
            "last": "2016-06-23T14:33:06-03:00",
            "name": "monitor",
            'backup_report': default_backup_report},
            {
            "last": "now",
            "name": "client1",
            "state": "idle",
            "phase": "phase1",
            "percent": 12,
            'backup_report': default_backup_report},
            {
            "last": "never",
            "name": "client2",
            "state": "idle",
            "phase": "phase2",
            "percent": 42,
            'backup_report': default_backup_report},
            {
                "last": 'now',
                "name": "cli30",
                "state": "idle",
                "phase": "null",
                "percent": 42,
                "server": 'server1',
                'backup_report': default_backup_report},
            {
                "last": actual_time.isoformat('T'),
                "name": "cli30",
                "state": "idle",
                "phase": "phase2",
                "percent": 0,
                "server": 'server2',
                'backup_report': default_backup_report}
        ]

        server = 'server1'
        # Add clients with random last in a range of 100
        for num in range(1, 101):
                    client_time = actual_time.replace(days=-random.randrange(5, 200))
                    client_last = client_time.isoformat(sep='T')
                    client_name = "client_{:03d}".format(num)
                    client_state = random.choice(['idle', 'working'])
                    client_phase = 'null'
                    client_percent = 0
                    if client_state == 'working':
                        client_phase = 'phase2'
                        client_percent = random.randrange(30, 90)
                    if num > 50:
                        server = 'server2'
                    
                    empty_backup_report = default_client_backup_report()
                    totsize = random.randrange(0, 10)
                    empty_backup_report['totsize'] = totsize

                    clients_stats.append({
                        "last": client_last,
                        "name": client_name,
                        "state": client_state,
                        "phase": client_phase,
                        "percent": client_percent,
                        "server": server,
                        'backup_report': empty_backup_report
                    })


        clients_stats[1]['backup_report']['totsize'] = 2048

        return clients_stats

    # @staticmethod
    # def get_backup_report():
    #     """
    #     dummy data example
    #     GET /api/client/(server)/report/(name)/(int: backup)
    #     GET /api/client/report/(name)/(int: backup)
    #
    #     :return:
    #     """
    #
    #     backup_report = [
    #         {"received": 326753806,
    #          "number": 1,
    #          "totsize": 572911431,
    #          "duration": 84
    #          }]



from ..lib.txt import TxtReports
import arrow
from collections import defaultdict
from ..lib.is_up import outdated_pings
from invpy_libs import csv_as_dict, save_csv_data, get_csv_from_url
from ..lib.email import EmailNotifications
import validators
from ..lib.files import temp_file
from ..lib.imap import ImapReceive


class BurpReports:
    def __init__(self, clients_dict, days_outdated=31, detail=None, config=None):
        """

        :param clients_dict: list of clients in burp_reports format, example:
        {'client_name':
            { 'b_last'    : 'YYYY-MM-DDTHH:mm:ssZZ',
              'b_state'    : 'working/current',
              'b_phase' : 'phase1/phase2',
              'b_last' : 'date',
            }
        }
        :param days_outdated (int): number to consider days outdated
        :param detail (True/None): Used to print more details or not
        :param config (configparser): configparser formatted config, required sections:
                            inventory_status, inventory_columns, common

        """

        self.clients = clients_dict
        self.days_outdated = days_outdated
        self.detail = detail
        self.config = config

        self.tempfile = temp_file('burp_reports_temp.csv')

        if config:
            # Set dict for from common section
            self.common_config = dict(self.config['common'])
            # Set dict for format_text section
            if config.has_section('format_text'):
                self.format_text = dict(self.config['format_text'])
            if config.has_section('inventory_columns'):
                self.inventory_columns = dict(self.config['inventory_columns'])
        else:
            self.common_config = {
                'csv_delimiter': ';'
            }

        self.excluded_clients = self.common_config.get('excluded_clients', '').split(',')

    def print_basic_txt(self):
        clients_reports = TxtReports(self.clients,
                                     detail=self.detail,
                                     format_text=self.format_text)
        clients_reports.report_to_txt()

    def _get_outdated(self):
        """

        :return: only outdated clients with b_status: outdated/never
        """

        outdated_clients = defaultdict(dict)

        for k, v in self.clients.items():
            # Get actual time with arrow module
            # k is client name
            # v is the dict with all data

            # Do not verify excluded clients
            if k in self.excluded_clients:
                continue

            actual_time = arrow.get()
            # get date to consider as outdated
            outdated_time = actual_time.replace(days=-self.days_outdated)
            client_last = v.get('b_last', None)

            # Add client to outdated list if not backup
            if client_last.lower() in 'never' or not client_last:
                outdated_clients[k] = v
                outdated_clients[k]['b_status'] = 'never'
                # Go to next client
                continue

            # Convert client_last to arrow date
            client_last = arrow.get(v.get('b_last', None))

            # Ensure client_last is not string now
            if not isinstance(client_last, str):
                # Add client to outdated list if outdated
                if client_last < outdated_time:
                    outdated_clients[k] = v
                    outdated_clients[k]['b_status'] = 'outdated'

            # When having details, check the totsize of the client)
            if self.detail:
                if v.get('backup_report').get('totsize') == 0:
                    outdated_clients[k]['b_status'] = 'never'

        return outdated_clients

    def compare_inventory(self, csv_inventory):
        """

        :param csv_inventory: Input filename to compare from (also can be an url to download it)
        :return: list csv_rows_inventory_status (status of each client) in nested list one
        row per client
        """

        # Set dict of status from config
        all_status = dict(self.config['inventory_status'])

        # Set variables of status as lists, also enable possibility to have more than one status:
        spare_status = all_status['spare'].split(',')
        active_status = all_status['active'].split(',')

        # Set dict of status from config
        all_columns = self.inventory_columns

        client_column = all_columns['client_name']
        delimiter = self.common_config['csv_delimiter']

        # Download the csv if it is in a url
        if validators.url(csv_inventory):
            get_csv_from_url(csv_inventory, csv_output=self.tempfile, delimiter=delimiter)
            csv_inventory = self.tempfile

        elif csv_inventory == 'email_inventory':
            if self.config.has_section('email_inventory'):
                email_inventory_config = dict(self.config['email_inventory'])
                email_inventory_obj = ImapReceive(config=email_inventory_config)
                email_inventory_obj.download_attachment() # Read email and download
                csv_inventory = email_inventory_obj.save_file
            else:
                print("you must setup the section [email_inventory]"
                      "in configuration to use email_inventory")
                exit(2)

        # Get inventory from CSV file
        inventory = csv_as_dict(csv_inventory, client_column,
                                delimiter=delimiter,
                                encoding=self.common_config.get('csv_encoding'))
        # Get a list of clients outdated in burp:
        outdated_clients = self._get_outdated()

        csv_rows_inventory_status = []
        headers = []

        headers.insert(0, client_column)  # Change First header column as client
        headers.insert(1, all_columns['status'])  # Change Second header column as status
        headers.insert(2, all_columns['server'])  # Change Third header column as server
        # Change Fourth header column as status detailed
        headers.insert(3, all_columns['sub_status'])

        first_client = next(iter(inventory.keys()))  # Gets first key of dict

        for column in sorted(inventory[first_client].keys()):  # Get headers from input inventory sorted
            if column.lower() not in all_columns.values():  # Add to the list only if not in the all_columns defined
                headers.append(column)

        # Prepare the header list to return:
        csv_rows_inventory_status.append(headers)  # First row as headers

        # Verify inventory and compare with clients in burp
        for k in sorted(inventory):
            client = k
            server_name = ''

            # Do not check excluded clients
            if client in self.excluded_clients:
                continue

            # Columns required to verify the status, lowercase and without spaces at end and beginning
            status = inventory[client].get(all_columns['status'], '').lower().strip()
            sub_status = inventory[client].get(all_columns['sub_status'], '').lower().strip()

            # If the client is in burp servers
            if client in self.clients:

                if sub_status in spare_status:
                    burp_status = all_status['spare_in_burp']

                elif status not in active_status:
                    burp_status = all_status['inactive_in_burp']

                else:
                    burp_status = all_status['in_inventory_updated']
                    # Set never/outdated if the client is outdated
                    if client in outdated_clients:
                        burp_status = outdated_clients[client]['b_status']

            # If the client is not in burp servers: ---
            # Define the status as ignored for clients spare
            elif sub_status in spare_status:
                burp_status = all_status['spare_not_in_burp']

            # Define ignored to client not active:
            elif status not in active_status:
                burp_status = all_status['inactive_not_in_burp']

            else:
                # Define absent in burp
                burp_status = all_status['in_inventory_not_in_burp']

            # Add server_name information
            if self.clients[client].get('server', None):
                server_name = self.clients[client]['server']
                # Mark the status of the client as duplicated if there is more than one server on it.
                if len(server_name) > 1:
                    burp_status = all_status['in_many_servers']

            # Generate list row with client's status and other data
            row = [client, burp_status, server_name, sub_status]

            # Add all other columns, starting from fourth column in headers
            for i in range(4, len(headers)):
                row.append(inventory[client].get(headers[i]))  # Insert each value of client to the row

            csv_rows_inventory_status.append(row)

        # Check if there is some client in burp but not in the inventory
        for burp_client in sorted(self.clients):
            # Do not check excluded clients
            if burp_client in self.excluded_clients:
                continue

            if burp_client not in inventory:
                burp_status = all_status['not_inventory_in_burp']
                if self.clients[burp_client].get('server', None):
                    server_name = self.clients[burp_client]['server']
                else:
                    server_name = []
                sub_status = ''
                # Generate list row with client's status and other data
                row = [burp_client, burp_status, server_name, sub_status]

                csv_rows_inventory_status.append(row)

        return csv_rows_inventory_status

    def save_compared_inventory(self, input_file, output_file):
        """

        :param input_file: csv inventory input
        :param output_file: csv filename output
        :return:
        """
        delimiter = self.common_config['csv_delimiter']

        rows_inventory_compared = self.compare_inventory(input_file)
        save_csv_data(rows_inventory_compared, output_file, csv_delimiter=delimiter)

    def report_outdated(self, export_txt=None, ping=None):
        """

        :param export_txt: use it if you want to print on screen, used with cli
        :param ping: Tries to ping the host to check it's availability
        :return: dict of outdated clients.
        """

        # Create a list of additional columns to add to the report
        columns = {'Status': 'b_status'}

        outdated_clients = self._get_outdated()

        if ping:
            # Check ping on each outdated client
            outdated_clients = outdated_pings(outdated_clients)
            # Add the column comments to report
            columns['Comments'] = 'comments'

        # Create the object to export the report
        clients_reports = TxtReports(outdated_clients,
                                     additional_columns=columns,
                                     detail=self.detail,
                                     format_text=self.format_text)

        if export_txt:
            clients_reports.report_to_txt()

        return outdated_clients

    def email_outdated(self):
        """
        Sends email based on its configuration file
        :return:
        """

        email = self._compose_email(report='outdated')
        return email.send_email()

    def _compose_email(self, report):
        """
        Composes emails

        :param report: (str): outdated to get a list of outdated clients
        :return: EmailNotifications class object
        """
        if self.config:
            if self.config.has_section('email_notification'):
                email_config = dict(self.config['email_notification'])
        else:
            raise SystemExit('email_notification section is required in config')

        if report == 'outdated':
            clients = self._get_outdated()

        else:
            clients = self.clients

        clients_reports = TxtReports(clients,
                                     detail=self.detail,
                                     foot_notes=email_config.get('foot_notes', ''),
                                     format_text=self.format_text)

        body_str = clients_reports.report_to_txt(print_text=None)

        email = EmailNotifications(email_config, body_str)

        return email
